// FILE: CSRTS8.CPP
// MODULE: LOAD
// PROGRAM: CSRTS
////////////////////////////////////////////////////////////////////////////////
//
//     Functions:
//             load_array()            reads STDIN (usually redirected) for lines
//                                     which it stores in main_array.  
//                                     then it parses the line into fields and calls
//                                     the field processors that are specified for
//                                     that field.  whole line justification and
//                                     case conversion are included inline.
//                                     the string to be used for sorting is stored
//                                     in sort_array.
//
//
////////////////////////////////////////////////////////////////////////////////

#include <fstream.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <alloc.h>


// * following define calculates the division of memory into the two arrays by
// * dividing the total memory by an approximation to the average number of characters
// * that will be stored for each line in the source.  the result is a number that can
// * be taken as the number of lines that can be stored.  that value is used as the
// * as the number of entries in the index array, even though doing so takes away from
// * the space available for string storage and invalidates the one of the initial assumptions
// * used to make the calculation.  this is ok since the other assumptions are almost
// * certainly wrong anyway.
#define max_index_def (long)(array_size / (chr_line + chr_sort))

#define max_string_len  134
// how_wide macro calculates total field width
#define how_wide(x) (fields_widths[x] + fields_decimal[x] + 1)

char c_8_notice[] = " Copyright 1991, Ted Davis\
                   Sort routine after Comb Sort,\
                   Copyright 1991, McGraw-Hill";
extern char string[82];                // a general purpose string variable.
extern char year[6];
extern char * str;                     // a pointer used mostly to shorten strings.
extern int start_col;                  // an index to the first column to be sorted.
extern int field_width;                // the number of characters to be sorted.
extern int flag;                       // the go-no-go flag.
extern int i, k, l, j;                 // general use integer variables
extern int mi, si;                     // indexes into the main and sort arrays.
extern int case_flag;                  // the "ignore case" flag.
extern int reverse_flag;               // the master reverse sort flag.
extern int justify_flag;               // right justification flag.
extern int left_flag;                  // left justification flag.
extern int blank_flag;                 // flag to force acceptance of blank lines.
extern int time_i;                     // index into the list of fields that are times.
extern int date_i;                     // index into the list of fields that are dates.
extern int Euro_flag;                  // marks dates as European (dd mm) sequence.
extern int reverse_i;                  // index into the list of field that are to be reverse sorted.
extern int case_i;                     // index into list of fields to be case insensitive.
extern int field_i;                    // index into the list of fields to be sorted.
extern int field_0;                    // flags that $0 was used to signal use of entire line.
extern int numbers_i;                  // index into numbers.
extern char * p;                       // another string pointer like strn.
extern int numbers[3];                 // temporary storage of command tail numbers.
extern int field_numbers[8];           // the list of fields to sort.
extern int fields_widths[8];           // the list of field widths.
extern int fields_decimal[8];          // the list of decimal field widths.
extern int time_fields[8];             // the list of fields that are times.
extern int date_fields[8];             // the list of fields that are dates.
extern int reverse_fields[8];          // the list of fields that are to be reverse sorted.
extern int case_fields[8];             // the list of fields that are to be case insensitive.
extern int number_fields[8];           // the list of fields that are to be sorted as numbers.
extern char fields[8][82];             // the fields to be sorted are stored here as strings.
extern char huge *main_array;          // * the string storage area, NOT actually a string array.
extern char huge *strn;                // * a transfer pointer, used for the string as read.
extern char huge *sort_str;            // * another transfer pointer, used for the sort pattern.
extern union                           // * the index array in two forms.
    {
        void far * prime_array;
        char huge *ix;
        char * huge *index_array;
    } ptr;
extern long array_size, long_i, long_x, long_j; // * limits and indexes.
extern long chr_line;                  // * used in division of memory into arrays.
extern long chr_sort;                  // * used in division of memory into arrays.
extern long max_index;                 // * used to limit array entries.




extern void time_convert(char *string, char *extra);
extern void date_convert(char *string, int flag, char *year);
extern void number_convert(char *string, int i_width, int d_width);
extern void reverse_convert(char *string);
extern void print_copyright_err(void);

int sl, blank_line;


void load_array(void);
void create_array(void);
void get_a_line(void);
void justify_right(char * string, int width);
void justify_left(char * string, int width);
void pad_right(char *field, int width);
void make_smallest(char *string, int i_width, int d_width);

void load_array(void)
// Reads lines from STDIN and puts them in main_array, justifies entire line or
// parses the line into fields, calls the conversion routines for individual
// field processing, then reassembles the sort line which is stored immediately
// following the source string.
{
int i, j;

    create_array();     // * divides free memory into a string storage area and an array of
                        // * huge pointers into the storage area.
    // array loader
        blank_line = 0;
    while((!flag) && (!cin.eof()))
    {   // * test for space to put string and pattern in array.
        if((long_j + max_string_len + max_string_len + 2) < array_size)
        {
            get_a_line();
        }
        else    // not enough room left.
        {
            flag = -1;
            print_copyright_err();
            cerr << "\nFATAL ERROR: NOT ENOUGH STRING SPACE\n";
            break;
        }
        if(l && (l < max_string_len))       // check string length.
        {
            if(!flag)            // if line is to be accepted and no error, so far.
            {
                // * following lines actually insert the strings into the array.
                if(long_i < (max_index - 1))  // * if there is room in the index array.
                {
                    if(!blank_line)
                    {
                        strcpy(strn, string);     // * this is a copy from a far string to a huge string.
                    }
                    else
                    {
                        strn[0] = '\0'; // record null string but leave pattern string alone.
                    }
                    ptr.index_array[long_i++] = strn;  // * insert source string.
                    ptr.index_array[long_i++] = sort_str;  // * insert sort pattern.
                // * sort_str does not yet exist but its starting point is now in the array.
                }
                else // out of index entries.
                {
                    print_copyright_err();
                    cerr << "\nFATAL ERROR: NOT ENOUGH TABLE ENTRIES\n";
                }
        // begin processing the sort pattern.

                str = string + start_col;   // point to starting column.
                strcpy(string, str);   // copy right hand part of string to string.
                // Note that that was a downward copy within the same string.
                if(!case_flag)   // if the entire sort is to be case insensitive
                {                // (the default).
                    strupr(string);      // convert to all upper case.
                }
                string[field_width] = '\0';     // truncate string to field_width.
        // full field case.
                if((field_i == -1) || (field_0))    //if using entire field or
                {                                   // a given field width.
                // right justify, if required, pad with spaces.
                    j = field_width - strlen(string);
                    if((j > 0) && justify_flag)
                    {
                        justify_right(string, field_width);

                    }
                    else    // left justify if that is required.
                    {
                        if((j > 0) && left_flag)
                        {
                            justify_left(string, field_width);
                        }
                    }
                    strcpy(sort_str, string);     // * copy sort pattern to the string storage area.
                    long_j += (strlen(sort_str) + 1);
                }
                else
        // use fields in a given order.
                {
                // first null the fields to prevent carryover, from previous, longer, lines.
                    fields[0][0]=fields[1][0]=fields[2][0]=fields[3][0]=\
                    fields[4][0]=fields[5][0]=fields[6][0]=fields[7][0] = '\0';
                // then split string into up to 8 fields.
                    sscanf(string,"%s%s%s%s%s%s%s%s",\
                    fields[0],fields[1],fields[2],fields[3],\
                    fields[4],fields[5],fields[6],fields[7]);
                // convert any that are times.
                    if(time_i != -1)
                    {
                        for(k = 0; k <= time_i; k++)
                        {
                            if(blank_line)
                            {
                                strcpy(fields[time_fields[k]], "00:00:00");
                            }
                            else
                            {
                                time_convert(fields[time_fields[k]],fields[time_fields[k] + 1]);
                            }
                        }
                    }
                // convert any that are dates.
                    if(date_i != -1)
                    {
                        for(k = 0; k <= date_i; k++)
                        {
                            if(blank_line)
                            {
                                strcpy(fields[date_fields[k]], "00000000");
                            }
                            else
                            {
                                date_convert(fields[date_fields[k]], Euro_flag, year);
                            }
                        }
                    }
                // convert any that are numbers.
                    if(numbers_i != -1)
                    {
                        for(k = 0; k <= numbers_i; k++)
                        {
                            if(blank_line)
                            {
                                make_smallest(fields[number_fields[k]], fields_widths[k], fields_decimal[k]);
                            }
                            else
                            {
                                number_convert(fields[number_fields[k]], fields_widths[k], fields_decimal[k]);
                            }
                        }
                    }
                // convert all except those that are to be case sensitive to all upper case.
                    if(case_i == -1)
                    {
                        for(k = 0; k <= case_i; k++)
                        {
                            str = strupr(fields[case_fields[k]]);
                            strcpy(fields[case_fields[k]], str);
                        }
                    }
                // convert any fields that are to be individually reverse sorted.
                    if(reverse_i != -1)
                    {
                        for(k = 0; k <= reverse_i; k++)
                        {
                            reverse_convert(fields[reverse_fields[k]]);
                        }
                    }
                // * then rearrange them in the given order and concatenate them,
                // padding selected fields out to specified width with spaces.
                    pad_right(fields[field_numbers[0]], how_wide(0));
                    strcpy(sort_str, fields[field_numbers[0]]);
                    sl = strlen(sort_str);
                    for(i = 1; i <= field_i; i++)
                    {
                        sl += strlen(fields[field_numbers[i]]);
                        if(sl <= max_string_len)
                        {
                            pad_right(fields[field_numbers[i]], how_wide(i));
                            strcat(sort_str, fields[field_numbers[i]]);
                        }
                        else    // composite line is too long
                        {
                            flag = -1;  // abort
                            print_copyright_err();
                            cerr << "\nFATAL ERROR: COMPOSITE OF SELECTED FIELDS EXCEEDS MAX LINE LENGTH\n";
                        }
                    }
                    long_j += (1 + strlen(sort_str)); // * point to next main string
                }
            }
            // reverse convert entire line, if required.
        }
        else
        {
            if(l == max_string_len) // getline cannot return a line longer than max_string_len
            {
                cerr << "\nLINE TOO LONG";
                flag = -1;
            }
        }
    }
}
void create_array(void)
// * The following block of code uses farmalloc to allocate ALL free memory to a
// * single large block, which is then split up into two arrays.  The first, and
// * smaller array is used to store pointers into the other (much larger) array.
// * Both of these arrays are huge; their size is determined by the amount of
// * memory available.  The larger array is used to store the strings read from
// * the source file and the strings used for the actual sort operation.  The
// * index array is made a specified fraction of the total memory, based on the
// * number of characters assumed to make up the average line from the source
// * and the length of the average sort pattern.  If the assumptions are exactly
// * correct, both arrays will fill up at the same time.
// * Even number entries in the index array are source strings, odd ones are the
// * sort patterns.  It is extremely difficult to get a pointer to a dynamically
// * allocated huge object.  I used a union of a void far * to accept the return
// * from farmalloc, a char huge * to transfer to the main array, and a
// * char * huge* for the index array.
{
    array_size = farcoreleft() - 0x400; // * this is the number of bytes available in free memory.
                                        // * leave a small pad.

    ptr.prime_array = farmalloc(array_size); // * allocate it all to the master array.
    if(ptr.prime_array == NULL)
    {
        print_copyright_err();
        cerr << "FATAL ERROR: UNABLE TO ALLOCATE MEMORY\n";
        abort();
    }
    max_index = max_index_def; // * see comments at #define above.
    long_j = max_index * 4; // * each index represents 4 bytes (or chars).
    main_array = ptr.ix + long_j;  // * starting point + four bytes per huge pointer.
    array_size -= long_j;    // * subtract the portion allocated to the index array.
    long_j = 0;  // * now long_j is the index into main_array.
}

void get_a_line(void)
// * Reads a line from STDIN into "string" after checking (above) that there is enough space
// * in the storage area for the worst possible case of line and pattern length.
// * The index array is not checked here.  The transfer pointers are set to the positions
// * in the storage area where the strings will go.  This routine can determine the length
// * of the main string, and therefore, can compute the starting point of the pattern;  it
// * cannot determine the length of the pattern, so it cannot calculate the next main string
// * position.  Replaces blank lines with 8 spaced character 1's.
{
    blank_line = 0;
    long_x = long_j;
    // * if there is enough room for the string and the sort pattern.
    cin.getline(string, (max_string_len + 1));   // read a line.
    l = strlen(string);
    if(!l || (strspn(string, " \t\n")))
    {
        strcpy(string, "\1 \1 \1 \1 \1 \1 \1 \1");
        blank_line = -1;
    }
    long_j += (l+ 1);
    if(blank_line) {l = 23;}
    strn = main_array + long_x;     // * set transfer pointer to the next string location.
    sort_str = main_array + long_j; // * set pattern transfer pointer to the one following that.
}

void justify_right(char * string, int width)
// Moves a string to the end of its allotted space and left fills with spaces.
{
int i, k;
    k = width + 1;
    for(i = strlen(string); i >= 0; i--)
    {
        string[k--] = string[i]; // note this moves the string to
    }                            // the right, including the terminator.
    for(j = k; j >= 0; j--)
    {
        string[j] = ' ';         // pad left from the first real character to 0.
    }
}

void justify_left(char * string, int width)
// Strips leading whitespace; moves the first string down so that the first non-whitespace
// character is in position 0; pads string out to "width" with spaces and terminates it.
{
int i = 0, k;
    while(isspace(string[i]))   // strip leading whitespace.
    {
        i++;
    }
    for(k = 0; string[i] != '\0'; k++)  // now at first good character.
    {
        string[k] = string[i++];    // move the characters down to the new positions.
    }
    for(i = k; i < width; i++)      // now, fill out to "width" with spaces.
    {
        string[i] = ' ';
    }
    string[i] = '\0';               // terminate string.
}

void pad_right(char *field, int width)
// Pads a string out to the specified width with spaces. Does nothing if the
// field is already as wide as width or wider (except reterminate it).
{
int i;
    for(i = strlen(field); i < width; i++)
    {
        field[i] = ' ';
    }
    field[i] = '\0';
}

void make_smallest(char *string, int i_width, int d_width)
// Returns the field as a string of character 1's a point and more 1's so
// that the string will sort as smaller than any possible number.
{
int i, j;
    j = i_width + d_width + 2;
    for(i = 0; i <= j; i++)
    {
        string[i] = '\1';
    }
    string[j] = '\0';
}

